"""
Classifies: CHEBI:19555 2-deoxy-D-glucoside
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_2_deoxy_D_glucoside(smiles: str):
    """
    Determines if a molecule is a 2-deoxy-D-glucoside compound.

    A 2-deoxy-D-glucoside compound is a D-glucoside compound with the 2-hydroxy substituent
    either absent or replaced by a different functional group.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is a 2-deoxy-D-glucoside, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Find the glucoside substructure
    glucoside_smarts = "OC1C(O)C(O)C(O)C(O)C1O"
    glucoside_mol = Chem.MolFromSmarts(glucoside_smarts)
    matches = mol.GetSubstructMatches(glucoside_mol)

    if not matches:
        return False, "Molecule does not contain a glucoside substructure"

    for match in matches:
        # Check if the 2-hydroxy substituent is absent or replaced
        atom_idx = match[1]
        atom = mol.GetAtomWithIdx(atom_idx)
        if atom.GetHybridization() == Chem.HybridizationType.SP3:
            if atom.GetNumExplicitHs() == 0:
                return True, "Molecule is a 2-deoxy-D-glucoside (2-hydroxy substituent absent)"
            elif atom.GetIsAromatic() or any(neighbor.GetIsAromatic() for neighbor in atom.GetNeighbors()):
                return True, "Molecule is a 2-deoxy-D-glucoside (2-hydroxy substituent replaced by an aromatic group)"
            else:
                substituent = ''.join(sorted([neighbor.GetSymbol() for neighbor in atom.GetNeighbors() if neighbor.GetIdx() not in match]))
                return True, f"Molecule is a 2-deoxy-D-glucoside (2-hydroxy substituent replaced by {substituent})"

    return False, "Molecule does not meet the criteria for a 2-deoxy-D-glucoside"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:19555',
                          'name': '2-deoxy-D-glucoside',
                          'definition': 'A D-glucoside compound with the '
                                        '2-hydroxy substituent either absent '
                                        'or replaced by a different functional '
                                        'group.',
                          'parents': ['CHEBI:35436']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': '\n'
               'Attempt failed: F1 score of 0 is too low.\n'
               'True positives: []\n'
               'False positives: '
               "[('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CC)=O)(OC(CCCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCCCC)COC(=O)CCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCCCCCC)COC(=O)CCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](O)COC(=O)CCCCCCCCC/C=C\\\\CCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@H](O)[C@@H](O)C(O)[C@@H](O)[C@H]1O)(OC[C@H](OC(=O)CCC/C=C\\\\[C@@H](O)/C=C/C=C\\\\C/C=C\\\\C=C\\\\[C@@H](O)C/C=C\\\\CC)COC(=O)CCCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCC/C=C\\\\CCCCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCC)COCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCC/C=C\\\\CCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O[C@@H]2OC([C@H](O)C(O)[C@@H]2O)CO)C(O)C(O)C(O)[C@H]1O[C@H]3OC([C@@H](O)C(O)[C@H]3O)CO)(OC[C@H](O)COC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCCCCC/C=C\\\\CCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCCCCCCCC)=O)(OC(CCCCCCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCCC)COC(=O)CCCCCCCCC/C=C\\\\CCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCCCCCCC)COC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@H](O)[C@H](O)C(O)[C@H](O)[C@H]1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCCC)([O-])=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@H](O)[C@@H](O)C(O)[C@H](O)[C@@H]1O)(O)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](O)COC(=O)CCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCCC)COCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('OC1C(O)C(O)C(OP([O-])(=O)OCC(CO[*])OC([*])=O)C(O)C1O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCC/C=C\\\\CCCCCC)=O)(OC(CCCCCCCCC/C=C\\\\CCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCC)COCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC(O)/C=C/C(O)=O)COC(=O)CCCCCCC/C=C\\\\CCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCC/C=C\\\\CCCCCCCC)COC(=O)CCCCCCC/C=C\\\\CCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@@H](O)[C@@H](O)C(OP(O)(O)=O)[C@H](O)[C@H]1O)(O)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCCCCCCCCCC)=O)(OC(CCCCCCC/C=C\\\\C/C=C\\\\CCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCCCCCCCC)=O)(OC(CCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('S(C[C@H](N)C(O[C@@H](COP(OC1[C@H](O)[C@H](O)C(O)[C@H](O)[C@H]1O)(O)=O)COC(=O)CCCCCCCCCCCCCCCCC)=O)[C@@H]([C@@H](O)CCCC(O)=O)/C=C/C=C/C=C\\\\C/C=C\\\\CCCCC', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)=O)(OC(CCCCCCC/C=C\\\\CCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCC/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCCC/C=C\\\\CCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCC/C=C\\\\CCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)COCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCCCC)COC(=O)CCCCCCC/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@H](O)[C@@H](O)C(O)[C@@H](O)[C@H]1O)(OCC(OC(=O)CCCCCCCCCCCCCCC)CO)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)C(O)C1O)(OCC(OC(=O)CCC/C=C\\\\CC2OC2C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCCC/C=C\\\\CCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('O(C1[C@H](O)[C@@H](O)C(OC(=O)CC=2C=3C(NC2)=CC=CC3)[C@@H](O)[C@@H]1O)[C@@H]4O[C@@H]([C@H](O)[C@H](O)[C@H]4O)CO', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCC/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('O(C1[C@@H](O)[C@H](O)C(O)[C@@H](O)[C@H]1O)C', 'Molecule is a "
               "2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCC/C=C\\\\CCCCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCCCCC)COCCCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCCCCCCCCCC)=O)(OC(CCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCC)CO/C=C\\\\CCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O[C@@H]2OC([C@H](O)C(O)[C@@H]2O)CO)C(O)C(O)C(O)[C@H]1O[C@H]3OC([C@@H](O)C(O)[C@H]3O)CO)(OC[C@H](O)COC(=O)CCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCCCCCCCCCC)=O)(OC(CCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCC/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCC/C=C\\\\CCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CC)=O)(OC(CCCCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('OC(CO[*])COP([O-])(=O)OC1C(O)C(O)C(O)C(O)C1O', 'Molecule is "
               "a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCC)COC(=O)CCCCCCCCC/C=C\\\\CCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCC/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCCCC/C=C\\\\CCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)COC(=O)CCCCCCC/C=C\\\\CCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCC)COC(=O)CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCC/C=C\\\\CCCCCC)=O)(OC(CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@H](O)[C@H](O)C(O)[C@H](O)[C@H]1O)(OC[C@H](O)CO)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O[C@@H]2OC([C@H](O)C(O)[C@@H]2O)CO)C(O)C(O)C(O)[C@H]1O[C@H]3OC([C@@H](O)C(O)[C@H]3O)CO)(OC[C@H](O)COC(=O)CCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCC)COCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCC)COC(=O)CCCCCCC/C=C\\\\CCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCC/C=C\\\\CCCCCC)=O)(OC(CCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](O)COC(=O)CCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCCCCCC)COC(=O)CCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1[C@H](O)[C@H](O)C(O)[C@H](O)[C@H]1O)(OC[C@H](OC(=O)CCCC(O)/C=C/C=C/C\\\\C=C\\\\CCCCCCCC)COC(=O)CCCCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCC(O)=O)COC(=O)CCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCCCCC)COC(=O)CCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\C/C=C\\\\CCCC)COC(=O)CCCCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCCC)COC(=O)CCCCCCC/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCC)COC(=O)CCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC(=O)/C=C/C=O)COC(=O)CCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCC)CO/C=C\\\\CCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCCCC)COC(=O)CCCCCCCCCCCCCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCC/C=C\\\\CCCCCC)=O)(OC(CCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCC/C=C\\\\CCCCCCC)COC(=O)CCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCC(=O)/C=C/C=O)COC(=O)CCCCCCC/C=C\\\\CCCCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('O(C1[C@H](O)[C@@H](O)C(O)[C@H](O)[C@@H]1O)[C@H]2O[C@@H]([C@H](O)[C@H](O)[C@H]2O)CO[C@H]3O[C@@H]([C@H](O)[C@H](O)[C@H]3O)CO[C@H]4O[C@@H]([C@H](O)[C@H](O)[C@H]4O)CO', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCC/C=C\\\\CCCCCCCC)=O)(OC(CCCCCCC/C=C\\\\CCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](O)COC(=O)CCCCCCC/C=C\\\\CCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CCCCCCCCCCCCCCC)=O)(OC(CCCCCCCC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('P(OC1C(O)C(O)C(O)[C@@H](O)C1O)(OC[C@H](OC(=O)CCCCCCCCCCC/C=C\\\\C/C=C\\\\CCCCC)COC(=O)CCCCCCC/C=C\\\\CCCCCC)(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside'), "
               "('[C@@H]1(C(C(C([C@H](C1O)O)O)O)O)OP(OC[C@@](COC(CC/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\C/C=C\\\\CCCCC)=O)(OC(CCCCCCCCCCCCCCC)=O)[H])(O)=O', "
               "'Molecule is a 2-deoxy-D-glucoside')]\n"
               'False negatives: '
               "[('CC(=O)N[C@@H](CSCC(=O)c1ccccc1)C(=O)N[C@@H]1[C@@H](O)[C@H](O)[C@@H](CO)O[C@@H]1O[C@H]1[C@H](O)[C@@H](O)[C@H](O)[C@@H](O)[C@H]1O', "
               "'Molecule contains 2-hydroxy substituent')]",
    'attempt': 1,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 1,
    'num_false_positives': 100,
    'num_true_negatives': 16470,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.009900990099009901,
    'recall': 1.0,
    'f1': 0.0196078431372549,
    'accuracy': 0.9939653611731338}


"""
Classifies: CHEBI:141498 hemiaminal ether
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_hemiaminal_ether(smiles: str):
    """
    Determines if a molecule is a hemiaminal ether.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if the molecule is a hemiaminal ether, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Find the nitrogen atoms
    nitrogen_atoms = [atom.GetIdx() for atom in mol.GetAtoms() if atom.GetSymbol() == 'N']

    # Check if there is at least one nitrogen atom
    if not nitrogen_atoms:
        return False, "No nitrogen atoms found"

    # Check if the nitrogen atom is connected to an oxygen and two carbons
    for nitrogen_idx in nitrogen_atoms:
        nitrogen_atom = mol.GetAtomWithIdx(nitrogen_idx)
        neighbors = [mol.GetAtomWithIdx(neighbor.GetIdx()) for neighbor in nitrogen_atom.GetNeighbors()]

        if len(neighbors) != 3:
            continue

        oxygen_neighbors = [neighbor for neighbor in neighbors if neighbor.GetSymbol() == 'O']
        carbon_neighbors = [neighbor for neighbor in neighbors if neighbor.GetSymbol() == 'C']

        if len(oxygen_neighbors) == 1 and len(carbon_neighbors) == 2:
            # Check if the oxygen is connected to a carbon
            oxygen_idx = oxygen_neighbors[0].GetIdx()
            oxygen_atom = mol.GetAtomWithIdx(oxygen_idx)
            oxygen_neighbors = [mol.GetAtomWithIdx(neighbor.GetIdx()) for neighbor in oxygen_atom.GetNeighbors()]
            carbon_neighbors_of_oxygen = [neighbor for neighbor in oxygen_neighbors if neighbor.GetSymbol() == 'C']

            if carbon_neighbors_of_oxygen:
                return True, "Molecule is a hemiaminal ether"

    return False, "Molecule is not a hemiaminal ether"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:141498',
                          'name': 'hemiaminal ether',
                          'definition': 'An organic amino compound that is a '
                                        'hemiaminal in which the hydrogen atom '
                                        'of the hydroxy group has been '
                                        'replaced by an organyl group. General '
                                        "formula: R2C(OR')NR2 ( R =/= H ). "
                                        'Also known as alpha-amino ethers.',
                          'parents': ['CHEBI:36963', 'CHEBI:50047']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 0,
    'num_false_positives': 100,
    'num_true_negatives': 52888,
    'num_false_negatives': 1,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0.0,
    'accuracy': 0.9980939440261186}


"""
Classifies: CHEBI:19168 17-oxo steroid
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_17_oxo_steroid(smiles: str):
    """
    Determines if a molecule is a 17-oxo steroid.
    
    Args:
        smiles (str): SMILES string of the molecule
        
    Returns:
        bool: True if molecule is a 17-oxo steroid, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None, "Invalid SMILES string"

    # Check for steroid core (4 fused rings)
    ri = mol.GetRingInfo()
    if len(ri.AtomRings()) < 4:
        return False, "Does not contain minimum 4 rings required for steroid core"
    
    # Get SMARTS pattern for steroid core with C17 ketone
    # Pattern matches 4 fused rings with C=O at position 17
    steroid_pattern = Chem.MolFromSmarts('[#6]~1~[#6]~[#6]~[#6]~2~[#6]~[#6]~[#6]~[#6]~3~[#6]~[#6]~[#6]~4~[#6](=O)~[#6]~[#6]~[#6]~4~[#6]~3~[#6]~2~1')
    
    if mol.HasSubstructMatch(steroid_pattern):
        # Find the C=O group at position 17
        matches = mol.GetSubstructMatches(steroid_pattern)
        
        if len(matches) > 0:
            return True, "Contains steroid core with C=O at position 17"
            
    return False, "Does not match 17-oxo steroid pattern"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:19168',
                          'name': '17-oxo steroid',
                          'definition': 'Any  oxo steroid carrying the oxo '
                                        'group at position 17.',
                          'parents': ['CHEBI:35789', 'CHEBI:3992']},
    'config': {   'llm_model_name': 'lbl/claude-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 0,
    'num_false_positives': 0,
    'num_true_negatives': 183860,
    'num_false_negatives': 7,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0.0,
    'accuracy': 0.9999619290030294}


"""
Classifies: CHEBI:27081 transition element atom
"""
from rdkit import Chem

# Define the atomic numbers for transition elements
TRANSITION_ELEMENT_ATOMIC_NUMBERS = list(range(21, 31)) + list(range(39, 49)) + list(range(71, 81)) + list(range(103, 113))

def is_transition_element_atom(smiles: str):
    """
    Determines if a molecule is a transition element atom.

    A transition element atom is defined as an element whose atom has an incomplete d sub-shell,
    or which can give rise to cations with an incomplete d sub-shell.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if the molecule is a transition element atom, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check if the molecule is a single atom
    if mol.GetNumAtoms() != 1:
        return False, "Molecule is not a single atom"

    atom = mol.GetAtomWithIdx(0)

    # Get the atomic number and symbol
    atomic_num = atom.GetAtomicNum()
    symbol = atom.GetSymbol()

    if atomic_num in TRANSITION_ELEMENT_ATOMIC_NUMBERS:
        return True, f"{symbol} is a transition element atom"

    return False, f"{symbol} is not a transition element atom"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:27081',
                          'name': 'transition element atom',
                          'definition': 'An element whose atom has an '
                                        'incomplete d sub-shell, or which can '
                                        'give rise to cations with an '
                                        'incomplete d sub-shell.',
                          'parents': ['CHEBI:33521']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': "Attempt failed: 'PeriodicGroup'",
    'attempt': 1,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 5,
    'num_false_positives': 68,
    'num_true_negatives': 183781,
    'num_false_negatives': 3,
    'num_negatives': None,
    'precision': 0.0684931506849315,
    'recall': 0.625,
    'f1': 0.12345679012345677,
    'accuracy': 0.9996138303137765}


"""
Classifies: CHEBI:140310 phenyl acetates
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_phenyl_acetates(smiles: str):
    """
    Determines if a molecule is a phenyl acetate - an acetate ester formed by condensation 
    of acetic acid with a phenol.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is a phenyl acetate, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check for presence of phenol acetate substructure
    phenyl_acetate_pattern = Chem.MolFromSmarts('c-OC(=O)C')
    if not mol.HasSubstructMatch(phenyl_acetate_pattern):
        return False, "No phenyl acetate group found"

    # Find all phenyl rings
    aromatic_rings = mol.GetAromaticAtoms()
    if not aromatic_rings:
        return False, "No aromatic rings found"

    # Get all matches of phenyl acetate pattern
    matches = mol.GetSubstructMatches(phenyl_acetate_pattern)
    
    # For each match, verify it's connected to a phenyl ring
    valid_matches = []
    for match in matches:
        phenoxy_oxygen = mol.GetAtomWithIdx(match[1])  # O atom in O-C(=O)C
        phenyl_carbon = mol.GetAtomWithIdx(match[0])   # aromatic C
        
        # Verify the carbon is part of aromatic ring
        if phenyl_carbon.GetIsAromatic():
            # Verify the oxygen is connected to aromatic carbon and acetyl group
            if (phenoxy_oxygen.GetDegree() == 2 and 
                any(n.GetIsAromatic() for n in phenoxy_oxygen.GetNeighbors())):
                valid_matches.append(match)

    if not valid_matches:
        return False, "No valid phenyl acetate groups found"

    # Count number of phenyl acetate groups
    num_acetates = len(valid_matches)
    
    # Get other substituents on the aromatic ring(s)
    other_substituents = []
    for atom in mol.GetAtoms():
        if atom.GetIsAromatic():
            for neighbor in atom.GetNeighbors():
                if not neighbor.GetIsAromatic() and neighbor.GetIdx() not in [m[1] for m in valid_matches]:
                    other_substituents.append(neighbor.GetSymbol())

    substituent_info = ""
    if other_substituents:
        substituent_info = f" with additional substituents: {', '.join(set(other_substituents))}"
    
    return True, f"Phenyl acetate with {num_acetates} acetate group(s){substituent_info}"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:140310',
                          'name': 'phenyl acetates',
                          'definition': 'An acetate ester obtained by formal '
                                        'condensation of the carboxy group of '
                                        'acetic acid with the hydroxy group of '
                                        'any phenol.',
                          'parents': ['CHEBI:22712', 'CHEBI:47622']},
    'config': {   'llm_model_name': 'lbl/claude-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 3,
    'num_false_positives': 100,
    'num_true_negatives': 33160,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.02912621359223301,
    'recall': 1.0,
    'f1': 0.05660377358490566,
    'accuracy': 0.9969936566154586}


"""
Classifies: CHEBI:2440 acrovestone
"""
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors

def is_acrovestone(smiles: str):
    """
    Determines if a molecule is an acrovestone - a polyphenol isolated from Acronychia pedunculata.
    
    Args:
        smiles (str): SMILES string of the molecule
        
    Returns:
        bool: True if molecule is an acrovestone, False otherwise
        str: Reason for classification
    """
    
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None, "Invalid SMILES string"
        
    # Check for presence of chromene/chromenone core
    chromene_pattern = Chem.MolFromSmarts('O1c2ccccc2C=CC1')
    chromenone_pattern = Chem.MolFromSmarts('O1c2ccccc2C(=O)C=C1')
    
    if not mol.HasSubstructMatch(chromene_pattern) and not mol.HasSubstructMatch(chromenone_pattern):
        return False, "Missing chromene/chromenone core structure"
        
    # Check for polyphenol characteristics
    oh_count = len(mol.GetSubstructMatches(Chem.MolFromSmarts('cOH')))
    if oh_count < 1:
        return False, "Missing phenolic OH groups"
        
    # Check for glycoside/glucuronide/malonate substituents
    glycoside_pattern = Chem.MolFromSmarts('OC1OC(CO)C(O)C(O)C1O')
    glucuronide_pattern = Chem.MolFromSmarts('OC1OC(C(=O)O)C(O)C(O)C1O')
    malonate_pattern = Chem.MolFromSmarts('OC(=O)CC(=O)')
    
    has_glycoside = mol.HasSubstructMatch(glycoside_pattern)
    has_glucuronide = mol.HasSubstructMatch(glucuronide_pattern)
    has_malonate = mol.HasSubstructMatch(malonate_pattern)
    
    if not (has_glycoside or has_glucuronide or has_malonate):
        return False, "Missing characteristic glycoside/glucuronide/malonate substituents"
    
    # Check molecular weight range
    mw = Descriptors.ExactMolWt(mol)
    if not (300 < mw < 1000):
        return False, "Molecular weight outside expected range"
        
    substituents = []
    if has_glycoside:
        substituents.append("glycoside")
    if has_glucuronide:
        substituents.append("glucuronide") 
    if has_malonate:
        substituents.append("malonate")
        
    return True, f"Acrovestone with {', '.join(substituents)} substituents"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:2440',
                          'name': 'acrovestone',
                          'definition': 'A polyphenol that is isolated from '
                                        'Acronychia pedunculata and exhibits '
                                        'moderate antioxidant and '
                                        'antityrosinase activities.',
                          'parents': [   'CHEBI:22187',
                                         'CHEBI:26195',
                                         'CHEBI:35618',
                                         'CHEBI:78840']},
    'config': {   'llm_model_name': 'lbl/claude-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': False,
    'best': True,
    'error': 'Python argument types in\n'
             '    Mol.GetSubstructMatches(Mol, NoneType)\n'
             'did not match C++ signature:\n'
             '    GetSubstructMatches(RDKit::ROMol self, RDKit::MolBundle '
             'query, RDKit::SubstructMatchParameters params)\n'
             '    GetSubstructMatches(RDKit::ROMol self, RDKit::ROMol query, '
             'RDKit::SubstructMatchParameters params)\n'
             '    GetSubstructMatches(RDKit::ROMol self, RDKit::MolBundle '
             'query, bool uniquify=True, bool useChirality=False, bool '
             'useQueryQueryMatches=False, unsigned int maxMatches=1000)\n'
             '    GetSubstructMatches(RDKit::ROMol self, RDKit::ROMol query, '
             'bool uniquify=True, bool useChirality=False, bool '
             'useQueryQueryMatches=False, unsigned int maxMatches=1000)',
    'stdout': '',
    'num_true_positives': 0,
    'num_false_positives': 0,
    'num_true_negatives': 0,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0.0,
    'accuracy': None}


"""
Classifies: CHEBI:33662 annulene
"""
from rdkit import Chem
from rdkit.Chem import Descriptors

def is_annulene(smiles: str):
    """
    Determines if a molecule is an annulene (mancude monocyclic hydrocarbon without side chains of the general formula CnHn (n is an even number) or CnHn+1 (n is an odd number)).

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is an annulene, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check if the molecule is a monocycle
    if not mol.GetRingInfo().IsAtomInRingOfSize(0, mol.GetNumAtoms()):
        return False, "Not a monocycle"

    # Check if the molecule contains only C and H atoms
    elements = {atom.GetSymbol() for atom in mol.GetAtoms()}
    if elements != {'C', 'H'}:
        return False, "Molecule contains atoms other than C and H"

    # Count the number of carbon and hydrogen atoms
    num_c = sum(atom.GetSymbol() == 'C' for atom in mol.GetAtoms())
    num_h = sum(atom.GetSymbol() == 'H' for atom in mol.GetAtoms())

    # Check if the molecule satisfies the general formula
    if num_c % 2 == 0 and num_h != num_c:
        return False, "Does not satisfy the general formula for even number of carbons"
    elif num_c % 2 == 1 and num_h != num_c + 1:
        return False, "Does not satisfy the general formula for odd number of carbons"

    # Check if the molecule has any substituents
    for atom in mol.GetAtoms():
        if atom.GetSymbol() == 'C' and atom.GetDegree() > 2:
            return False, "Molecule contains substituents"

    return True, f"Annulene with {num_c} carbon atoms"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:33662',
                          'name': 'annulene',
                          'definition': 'A mancude monocyclic hydrocarbon '
                                        'without side chains of the general '
                                        'formula CnHn (n is an even number) or '
                                        'CnHn+1 (n is an odd number). In '
                                        'systematic nomenclature an annulene '
                                        'with seven or more carbon atoms may '
                                        'be named [n]annulene, where n is the '
                                        'number of carbon atoms.',
                          'parents': ['CHEBI:33664']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 0,
    'num_false_positives': 0,
    'num_true_negatives': 183925,
    'num_false_negatives': 1,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0,
    'accuracy': 0.9999945630307842}


"""
Classifies: CHEBI:26004 phenylpropanoid
"""
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors

def is_phenylpropanoid(smiles: str):
    """
    Determines if a molecule is a phenylpropanoid based on its structure.
    
    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is a phenylpropanoid, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check for aromatic ring
    rings = mol.GetRingInfo()
    has_aromatic_ring = False
    for ring in rings.AtomRings():
        if len(ring) == 6:
            atoms = [mol.GetAtomWithIdx(i) for i in ring]
            if all(atom.GetIsAromatic() for atom in atoms):
                has_aromatic_ring = True
                break
    
    if not has_aromatic_ring:
        return False, "No aromatic ring found"

    # Look for phenylpropane skeleton (C6-C3)
    matches = mol.GetSubstructMatches(Chem.MolFromSmarts('c1ccccc1CCC'))
    if not matches:
        # Also check for modified phenylpropane structures common in phenylpropanoids
        patterns = [
            'c1ccccc1CC=C',  # Propenyl chain
            'c1ccccc1C(=O)CC', # Ketone
            'c1ccccc1COC', # Methoxy
            'c1ccccc1C1OCC1', # Cyclic ether
            'c1ccccc1C=CC=O', # Cinnamaldehyde
            'c1ccccc1C=CC(=O)O', # Cinnamic acid
            'c1ccccc1CC(=O)O', # Phenylacetic acid
        ]
        
        found_pattern = False
        for pattern in patterns:
            if mol.HasSubstructMatch(Chem.MolFromSmarts(pattern)):
                found_pattern = True
                break
                
        if not found_pattern:
            return False, "No phenylpropane or related skeleton found"

    # Check for common phenylpropanoid functionalities
    functionalities = []
    if mol.HasSubstructMatch(Chem.MolFromSmarts('cO')):
        functionalities.append("phenolic hydroxyl")
    if mol.HasSubstructMatch(Chem.MolFromSmarts('cOC')):
        functionalities.append("methoxy")
    if mol.HasSubstructMatch(Chem.MolFromSmarts('C(=O)O')):
        functionalities.append("carboxylic acid")
    if mol.HasSubstructMatch(Chem.MolFromSmarts('C=O')):
        functionalities.append("carbonyl")
    
    if functionalities:
        return True, f"Phenylpropanoid with {', '.join(functionalities)} groups"
    else:
        return True, "Basic phenylpropanoid structure"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:26004',
                          'name': 'phenylpropanoid',
                          'definition': 'Any organic aromatic compound with a '
                                        'structure based on a phenylpropane '
                                        'skeleton. The class includes '
                                        'naturally occurring phenylpropanoid '
                                        'esters, flavonoids, anthocyanins, '
                                        'coumarins and many small phenolic '
                                        'molecules as well as their '
                                        'semi-synthetic and synthetic '
                                        'analogues. Phenylpropanoids are also '
                                        'precursors of lignin.',
                          'parents': ['CHEBI:33659']},
    'config': {   'llm_model_name': 'lbl/claude-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 217,
    'num_false_positives': 100,
    'num_true_negatives': 423,
    'num_false_negatives': 96,
    'num_negatives': None,
    'precision': 0.6845425867507886,
    'recall': 0.6932907348242812,
    'f1': 0.6888888888888889,
    'accuracy': 0.7655502392344498}


"""
Classifies: CHEBI:139340 1,3-dihydroimidazole-2-thiones
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_1_3_dihydroimidazole_2_thiones(smiles: str):
    """
    Determines if a molecule is a 1,3-dihydroimidazole-2-thione or its derivative.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is a 1,3-dihydroimidazole-2-thione or its derivative, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check if the molecule contains an imidazole ring
    imidazole_ring = AllChem.FindMolRingOfSize(mol, 5)
    if not imidazole_ring:
        return False, "No imidazole ring found"

    # Check if the imidazole ring contains a sulfur atom
    s_atom_idx = [atom.GetIdx() for atom in mol.GetAtoms() if atom.GetSymbol() == 'S']
    if not any(s_atom_idx in ring for ring in mol.GetRingInfo().AtomRings()):
        return False, "No sulfur atom in the imidazole ring"

    # Check if the sulfur atom is connected to a carbon atom
    s_atom = mol.GetAtomWithIdx(s_atom_idx[0])
    if not any(neighbor.GetSymbol() == 'C' for neighbor in s_atom.GetNeighbors()):
        return False, "Sulfur atom not connected to a carbon atom"

    # Check for substitution
    substituents = []
    for atom in mol.GetAtoms():
        if atom.GetSymbol() != 'C' and atom.GetSymbol() != 'N' and atom.GetSymbol() != 'S':
            substituents.append(atom.GetSymbol())

    if substituents:
        substituents_str = ', '.join(set(substituents))
        return True, f"1,3-dihydroimidazole-2-thione with substituents: {substituents_str}"
    else:
        return True, "Unsubstituted 1,3-dihydroimidazole-2-thione"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:139340',
                          'name': '1,3-dihydroimidazole-2-thiones',
                          'definition': 'A member of the class of imidazoles '
                                        'that is 1,3-dihydroimidazole-2-thione '
                                        'and its derivatives by substitution.',
                          'parents': ['CHEBI:24780', 'CHEBI:51276']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': False,
    'best': True,
    'error': "module 'rdkit.Chem.AllChem' has no attribute 'FindMolRingOfSize'",
    'stdout': '',
    'num_true_positives': 0,
    'num_false_positives': 0,
    'num_true_negatives': 0,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0.0,
    'accuracy': None}


"""
Classifies: CHEBI:22216 acrylamides
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_acrylamides(smiles: str):
    """
    Determines if a molecule is an acrylamide or a derivative of acrylamide.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is an acrylamide, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Find all amide bonds
    amide_bonds = []
    for bond in mol.GetBonds():
        begin_atom = mol.GetAtomWithIdx(bond.GetBeginAtomIdx())
        end_atom = mol.GetAtomWithIdx(bond.GetEndAtomIdx())
        if begin_atom.GetSymbol() == 'N' and end_atom.GetSymbol() == 'C' and begin_atom.GetIsAromatic() is False and end_atom.GetIsAromatic() is False:
            if begin_atom.GetTotalNumHs() == 0 and end_atom.GetTotalNumHs() == 0:
                amide_bonds.append(bond.GetBeginAtomIdx())
                amide_bonds.append(bond.GetEndAtomIdx())

    if not amide_bonds:
        return False, "No amide bonds found"

    # Find all alkene bonds
    alkene_bonds = []
    for bond in mol.GetBonds():
        if bond.GetBondType() == Chem.BondType.DOUBLE:
            begin_atom = mol.GetAtomWithIdx(bond.GetBeginAtomIdx())
            end_atom = mol.GetAtomWithIdx(bond.GetEndAtomIdx())
            if begin_atom.GetHybridization() == Chem.HybridizationType.SP2 and end_atom.GetHybridization() == Chem.HybridizationType.SP2:
                alkene_bonds.append(bond.GetBeginAtomIdx())
                alkene_bonds.append(bond.GetEndAtomIdx())

    if not alkene_bonds:
        return False, "No alkene bonds found"

    # Check if any alkene is adjacent to an amide
    for alkene_idx in alkene_bonds:
        alkene_atom = mol.GetAtomWithIdx(alkene_idx)
        for neighbor in alkene_atom.GetNeighbors():
            if neighbor.GetIdx() in amide_bonds:
                return True, "Molecule is an acrylamide"

    return False, "Molecule is not an acrylamide"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:22216',
                          'name': 'acrylamides',
                          'definition': 'An enamide which is acrylamide or a '
                                        'derivative of acrylamide obtained by '
                                        'replacement of one or more of its '
                                        'hydrogens.',
                          'parents': ['CHEBI:51751']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': "Attempt failed: type object 'BondType' has no attribute "
               "'AMIDE'",
    'attempt': 2,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 1,
    'num_false_positives': 100,
    'num_true_negatives': 686,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.009900990099009901,
    'recall': 1.0,
    'f1': 0.0196078431372549,
    'accuracy': 0.8729351969504447}


"""
Classifies: CHEBI:24059 fluorenes
"""
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem.AllChem import GetSymmSSSR

def is_fluorenes(smiles: str):
    """
    Determines if a molecule is a fluorene derivative.
    Fluorenes contain two benzene rings ortho-fused to a cyclopentane ring.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is a fluorene, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None, "Invalid SMILES string"

    # Get ring information
    rings = list(GetSymmSSSR(mol))
    if len(rings) < 3:
        return False, "Less than 3 rings found - fluorenes require 3 fused rings"

    # Find 5-membered rings
    five_rings = []
    for ring in rings:
        if len(ring) == 5:
            five_rings.append(ring)

    # Find 6-membered aromatic rings
    six_rings = []
    for ring in rings:
        if len(ring) == 6:
            atoms = [mol.GetAtomWithIdx(i) for i in ring]
            if all(atom.GetIsAromatic() for atom in atoms):
                six_rings.append(ring)

    if len(six_rings) < 2:
        return False, "Less than 2 benzene rings found"
    
    if len(five_rings) < 1:
        return False, "No 5-membered ring found"

    # Check if rings are properly fused
    for five_ring in five_rings:
        benzene_count = 0
        for six_ring in six_rings:
            # Check if rings share exactly 2 atoms (ortho-fusion)
            shared_atoms = set(five_ring).intersection(set(six_ring))
            if len(shared_atoms) == 2:
                benzene_count += 1
        
        if benzene_count == 2:
            # Found a 5-membered ring fused to 2 benzene rings
            return True, "Contains fluorene core structure (two benzene rings ortho-fused to cyclopentane)"

    return False, "Rings are not properly fused in fluorene arrangement"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:24059',
                          'name': 'fluorenes',
                          'definition': 'An ortho-fused polycyclic arene in '
                                        'which the skeleton is composed of two '
                                        'benzene rings ortho-fused to '
                                        'cyclopentane.',
                          'parents': ['CHEBI:38032']},
    'config': {   'llm_model_name': 'lbl/claude-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 7,
    'num_false_positives': 100,
    'num_true_negatives': 15885,
    'num_false_negatives': 3,
    'num_negatives': None,
    'precision': 0.06542056074766354,
    'recall': 0.7,
    'f1': 0.11965811965811965,
    'accuracy': 0.9935604876523914}


from pathlib import Path
from typing import Tuple, Iterator

import pandas as pd
from oaklib.cli import ontology_versions
from oaklib.datamodels.vocabulary import HAS_DEFINITION_CURIE, RDFS_LABEL, OWL_VERSION_IRI
from rdflib.plugins.shared.jsonld.keys import VERSION
from rdkit import Chem
from semsql.sqla.semsql import Statements, Edge, EntailedEdge
from sqlalchemy import select, Select, not_
from sqlalchemy.orm import Session
from sqlalchemy.orm.loading import instances
from sssom.constants import RDFS_SUBCLASS_OF

from c3p.datamodel import ChemicalStructure, ChemicalClass, Dataset

SMILES = "obo:chebi/smiles"

import re


def sanitize_smiles(smiles_string):
    """
    Sanitizes a SMILES string by:
    1. Removing whitespace
    2. Removing invalid characters
    3. Preserving valid SMILES characters including brackets, numbers, and symbols

    Args:
        smiles_string (str): Input SMILES string

    Returns:
        str: Sanitized SMILES string
    """
    # Remove whitespace
    smiles = smiles_string.strip()

    # Define valid SMILES characters
    # Includes:
    # - Atomic symbols (B, C, N, O, P, S, F, Cl, Br, I, etc.)
    # - Numbers and % for ring closures
    # - Special characters ([, ], (, ), =, #, /, \, @, +, -, ., *)
    # - Colons for aromatic bonds
    # - Commas for atom lists in brackets
    pattern = r'[^A-Za-z0-9\[\]\(\)=#/\\@+\-\.\*%:,]'

    # Remove invalid characters
    sanitized = re.sub(pattern, '', smiles)

    return sanitized

def eav_to_df(eav_df: pd.DataFrame, value_column='value') -> pd.DataFrame:
    """
    Convert an EAV DataFrame to a wide-format DataFrame.

    Example:

        >>> test_eav_df = pd.DataFrame({
        ...    'subject': ['a', 'a', 'b', 'b', 'b', 'c'],
        ...    'predicate': ['p1', 'p1', 'p1', 'p1', 'p2', 'p3'],
        ...    'value': ['v1', 'v2', 'v3', 'v4', 'v4', '']
        ... })
        >>> eav_to_df(test_eav_df)
        predicate subject        p1    p2    p3
        0               a  [v1, v2]  None  None
        1               b  [v3, v4]    v4  None
        2               c      None  None

    Args:
        eav_df:
        value_column:

    Returns:

    """
    # For predicates that have multiple values per subject,
    # aggregate them into lists
    unmasked_df = (eav_df.groupby(['subject', 'predicate'])[value_column]
                   .agg(list)
                   .unstack(fill_value=[])
                   .reset_index())

    # If we know some predicates should be single-valued,
    # we can unwrap them from lists
    single_value_mask = unmasked_df.apply(lambda x: x.map(len) <= 1)
    df = unmasked_df.mask(single_value_mask, unmasked_df.apply(lambda x: x.map(lambda y: y[0] if y else None)))

    return df

def db_to_dataframe(session: Session, prefix = "CHEBI") -> pd.DataFrame:
    """
    Convert a semsql database to a DataFrame

    Args:
        session:
        prefix:

    Returns:

    """
    def _filter(q: Select, tbl=Statements) -> Select:
        if prefix:
            q = q.where(tbl.subject.startswith(prefix))
        else:
            q = q.where(not_(tbl.subject.startswith("_:")))
        return q
    # Data triples
    q = select(Statements.subject,
                  Statements.predicate,
                  Statements.value).where(Statements.value != None)
    q = _filter(q)
    triples = pd.DataFrame(session.execute(q).all())
    v_df = eav_to_df(triples)
    # Object triples
    q = select(Statements.subject,
               Statements.predicate,
               Statements.object).where(Statements.object != None)
    q = q.where(not_(Statements.predicate == RDFS_SUBCLASS_OF))
    q = _filter(q)
    triples = pd.DataFrame(session.execute(q).all())
    o_df = eav_to_df(triples, value_column="object")
    # Edge triples
    q = select(Edge.subject,
               Edge.predicate,
               Edge.object).where(not_(Edge.object.startswith("_:")))
    q = _filter(q, Edge)
    triples = pd.DataFrame(session.execute(q).all())
    e_df = eav_to_df(triples, value_column="object")
    # Ancestor triples
    q = select(EntailedEdge.subject,
               EntailedEdge.object).where(EntailedEdge.predicate == RDFS_SUBCLASS_OF)
    q = q.where(not_(EntailedEdge.object.startswith("_:")))
    q = _filter(q, EntailedEdge)
    triples = pd.DataFrame(session.execute(q).all())
    triples['predicate'] = "entailed_subclass_of"
    a_df = eav_to_df(triples, value_column="object")
    df = v_df.merge(o_df, on='subject', how='outer').merge(e_df, on='subject', how='outer').merge(a_df, on='subject',
                                                                                                  how='outer')
    return df


def get_ontology_version(session: Session):
    q = select(Statements.object).where(Statements.predicate == OWL_VERSION_IRI)
    versions = session.execute(q).all()
    if len(versions) != 1:
        raise ValueError(f"Expected one version, got {len(versions)}")
    v = versions[0][0]
    # include ontology
    ont = "chebi"
    if "chemessence" in v:
        ont = "chemessence"
    return ont + v.split("/")[-2]

def create_benchmark(df: pd.DataFrame, session: Session, min_members=50, max_members=5000) -> Dataset:
    v = get_ontology_version(session)
    structure_df = df[df[SMILES].notnull()]
    defined_df = df[df[HAS_DEFINITION_CURIE].notnull()]
    cc_map = {}
    for _, row in defined_df.iterrows():
        parents = row[RDFS_SUBCLASS_OF]
        if parents and not isinstance(parents, list):
            if isinstance(parents, str):
                parents = [parents]
            else:
                parents = []
        cc = ChemicalClass(id=row["subject"], name=row[RDFS_LABEL], definition=row[HAS_DEFINITION_CURIE],
                           parents=parents,
                           instances=[], negative_instances=[])
        cc_map[cc.id] = cc
    s_map = {}
    for _, row in structure_df.iterrows():
        smiles_str = row[SMILES]
        #smiles_str_sanitized = sanitize_smiles(smiles_str)
        #if smiles_str_sanitized != smiles_str:
        #    raise ValueError(f"Invalid SMILES string: {smiles_str} != {smiles_str_sanitized}")
        structure = ChemicalStructure(name=row[RDFS_LABEL], smiles=smiles_str)
        s_map[row[SMILES]] = structure
        for a in row["entailed_subclass_of"]:
            if a == row["subject"]:
                continue
            if a in cc_map:
                cc = cc_map[a]
                cc.instances.append(structure)
                if len(cc.instances) > max_members:
                    del cc_map[a]
    ccs = [v for v in cc_map.values() if len(v.instances) >= min_members]
    if False:
        # TODO: calculate negatives later as we want to include all
        for cc in ccs:
            negative_smiles = []
            positive_smiles = [s.smiles for s in cc.instances]
            num_positive = len(positive_smiles)
            negative_candidates = list(set(s_map.keys()).difference(positive_smiles))
            parents = cc.parents
            for p in parents:
                if p in cc_map:
                    for i in cc_map[p].instances:
                        s = i.smiles
                        if s not in positive_smiles:
                            negative_smiles.append(s)
            negative_smiles = negative_smiles[:num_positive]
            if len(negative_smiles) < num_positive:
                import random
                random.shuffle(negative_candidates)
                negative_smiles.extend(negative_candidates[:num_positive - len(negative_smiles)])
            cc.negative_instances = [s_map[s] for s in negative_smiles]
    return Dataset(
        ontology_version=v,
        min_members=min_members,
        max_members=max_members,
        classes=ccs,
        structures=list(s_map.values()),
    )

def validate_dataset(dataset: Dataset) -> Iterator[Tuple[ChemicalStructure, str]]:
    """
    Validate a dataset

    Args:
        dataset:

    Returns:

    """
    for s in dataset.structures:
        smiles_str = s.smiles
        smiles_str_sanitized = sanitize_smiles(smiles_str)
        if smiles_str_sanitized != smiles_str:
            yield s, smiles_str_sanitized
        _mol = Chem.MolFromSmiles(smiles_str)

def create_and_save_benchmark(*args, **kwargs) -> Dataset:
    benchmark = create_benchmark(*args, **kwargs)
    benchmarks_dir = Path("inputs")
    benchmarks_dir.mkdir(exist_ok=True)
    path = f"{benchmarks_dir / benchmark.name}.json"
    with open(path, "w") as f:
        f.write(benchmark.model_dump_json(indent=2))
    return benchmark









"""
Classifies: CHEBI:137040 short chain fatty acyl-CoA(4-)
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_short_chain_fatty_acyl_CoA_4__(smiles: str):
    """
    Determines if a molecule is a short chain fatty acyl-CoA(4-), defined as 'A fatty acyl-CoA(4-) arising from deprotonation
    of the phosphate and diphosphate OH groups of any short-chain fatty acyl-CoA; major species at pH 7.3'.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if the molecule is a short chain fatty acyl-CoA(4-), False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check for the presence of CoA moiety
    coa_pattern = Chem.MolFromSmarts('C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12')
    match = mol.HasSubstructMatch(coa_pattern)
    if not match:
        return False, "CoA moiety not found"

    # Check for the presence of a short chain fatty acyl group
    fatty_acyl_pattern = Chem.MolFromSmarts('C(=O)C[C;H1]')
    match = mol.HasSubstructMatch(fatty_acyl_pattern)
    if not match:
        return False, "Short chain fatty acyl group not found"

    # Check for the presence of four negative charges
    charge = AllChem.GetFormalCharge(mol)
    if charge != -4:
        return False, f"Incorrect charge ({charge}), should be -4"

    return True, "Molecule is a short chain fatty acyl-CoA(4-)"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:137040',
                          'name': 'short chain fatty acyl-CoA(4-)',
                          'definition': 'A fatty acyl-CoA(4-) arising from '
                                        'deprotonation of the phosphate and '
                                        'diphosphate OH groups of any '
                                        'short-chain fatty acyl-CoA; major '
                                        'species at pH 7.3.',
                          'parents': ['CHEBI:77636']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.8,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': '\n'
               'Attempt failed: F1 score of 0 is too low.\n'
               'True positives: []\n'
               'False positives: '
               "[('[H][C@@](C)(CCC(=O)C(C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@@]1([H])CC[C@@]2([H])[C@]3([H])[C@H](O)C[C@]4([H])C[C@H](O)CC[C@]4(C)[C@@]3([H])C[C@H](O)[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC(C)(COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@@H](O)C(=O)NCCC(=O)NCCSC(=O)CC[C@H]1[C@@H]2CCC(=O)[C@@]2(C)CCC1=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)CC[C@@](C)([C@]4(CC[C@@]5([C@@]4(CC[C@@]6([C@]7(CCC(C=C7C=C[C@@]56[H])=O)C)[H])C)[H])[H])[H])=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H]([C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@H](CCC(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC(C(C)=O)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC(C)(COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@@H](O)C(=O)NCCC(=O)NCCSC(=O)CC(=O)C=Cc1ccc(O)cc1', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)CC[C@@](C)([C@]4(CC[C@@]5([C@@]4([C@H](C[C@@]6([C@]7(CCC(C=C7CC[C@@]56[H])=O)C)[H])O)C)[H])[H])[H])=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)CC[C@@](C)([C@]4(CC[C@@]5([C@@]4(CC[C@@]6([C@]7(CCC(C=C7C[C@H]([C@@]56[H])O)=O)C)[H])C)[H])[H])[H])=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H](O[C@@H]1O[C@@H](C)[C@H](O)C[C@H]1O)C(=O)CC(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)[C@@H]2OP([O-])([O-])=O)N3C=NC4=C(N)N=CN=C34', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C(C[C@@H]([C@]([C@@]1([C@]2(CC[C@@]3([C@]4(CCC(C=C4CC[C@]3([C@@]2(CC1)[H])[H])=O)C)[H])C)[H])(C)[H])O)(=O)SCCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]5O[C@@H](N6C7=C(C(=NC=N7)N)N=C6)[C@@H]([C@@H]5OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@H](C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)C1(O)CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C12=C(N=CN=C1N)N(C=N2)[C@@H]3O[C@@H]([C@@H](C3O)OP(=O)([O-])[O-])COP(=O)(OP(=O)(OCC([C@](C(NCCC(NCCSC(CCCCCCC[C@@H]4[C@@H](C(CC4)=O)C/C=C\\\\CC)=O)=O)=O)(O)[H])(C)C)[O-])[O-]', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC(C)(COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@@H](O)C(=O)NCCC(=O)NCCSC(=O)C1C(O)CCCC1=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('S(C(=O)CCCCC[C@@H]1[C@@H](C(C=C1)=O)C/C=C\\\\CC)CCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]2O[C@@H](N3C4=C(C(=NC=N4)N)N=C3)[C@@H]([C@@H]2OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@H](CCC[C@H](C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)CC[C@@](C)([C@]4(CC[C@@]5([C@@]4(CC[C@@]6([C@]7(CCC(C=C7C[C@@H]([C@@]56[H])O)=O)C)[H])C)[H])[H])[H])=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H]([C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)C=C[C@]4(C)[C@@]3(O)CC[C@]12C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC(C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)=C1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(CCC/C=C\\\\C/C=C\\\\CC(/C=C/C=C\\\\CCCCC)=O)=O)=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(C(C(CC\\\\C(=C\\\\CC4=C(C(=C5C(=C4O)C(OC5)=O)C)OC)\\\\C)=O)C)=O)=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@H](CCC=C(C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H]([C@H]1CC[C@H]2[C@H](CCc3cc(O)ccc3C)C(=O)CC[C@]12C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@H](C=CC(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C(/C=C/[C@]([C@@]1([C@]2(CC[C@@]3([C@]4(CCC(C=C4CC[C@]3([C@@]2(CC1)[H])[H])=O)C)[H])C)[H])(C)[H])(=O)SCCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]5O[C@@H](N6C7=C(C(=NC=N7)N)N=C6)[C@@H]([C@@H]5OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C(C(NCCC(NCCSC(=O)CC(=O)/C=C/C4=CC=C(C(=C4)OC)O)=O)=O)O)(C)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(CCC/C=C\\\\C/C=C\\\\C\\\\C=C/C=C/C(CCCCC)=O)=O)=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CCCCCCCCCCCCCC(=O)C(C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H]([C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)C=C[C@]4(C)[C@H]3CC[C@]12C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('COc1cc(C=CC(=O)CC(=O)SCCNC(=O)CCNC(=O)C(O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)[C@@H]2OP([O-])([O-])=O)n2cnc3c(N)ncnc23)ccc1O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C(CC[C@]([C@@]1([C@]2([C@H](C[C@@]3([C@]4(CCC(C=C4C=C[C@]3([C@@]2(CC1)[H])[H])=O)C)[H])O)C)[H])(C)[H])(=O)SCCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]5O[C@@H](N6C7=C(C(=NC=N7)N)N=C6)[C@@H]([C@@H]5OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('S(C(CCCCCCC[C@@H]1[C@@H](C(C=C1)=O)C/C=C\\\\CC)=O)CCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]2O[C@@H](N3C4=C(C(=NC=N4)N)N=C3)[C@@H]([C@@H]2OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)C(C)C(CC[C@]([C@@]4([C@]5(CC[C@@]6([C@]7(CC[C@H](C[C@]7(C[C@H]([C@]6([C@@]5(CC4)[H])[H])O)[H])O)C)[H])C)[H])(C)[H])=O)=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('S(C(=O)CCCCC[C@@H]1[C@@H](C(CC1)=O)C/C=C\\\\CC)CCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]2O[C@@H](N3C4=C(C(=NC=N4)N)N=C3)[C@@H]([C@@H]2OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C(C[C@H]([C@]([C@@]1([C@]2(CC[C@@]3([C@]4(CCC(C=C4CC[C@]3([C@@]2(CC1)[H])[H])=O)C)[H])C)[H])(C)[H])O)(=O)SCCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]5O[C@@H](N6C7=C(C(=NC=N7)N)N=C6)[C@@H]([C@@H]5OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC(C)CCCC(C)CCCC(C)CCC(=O)C(C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)[C@H](C)CCC[C@]([C@@]4([C@]5(CC[C@@]6([C@]7(CCC(C=C7CC[C@]6([C@@]5(CC4)[H])[H])=O)C)[H])C)[H])(C)[H])=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C(CC[C@]([C@@]1([C@]2([C@H](C[C@@]3([C@]4(CCC(C=C4C[C@H]([C@]3([C@@]2(CC1)[H])[H])O)=O)C)[H])O)C)[H])(C)[H])(=O)SCCNC(CCNC(=O)[C@@H](C(COP(OP(OC[C@H]5O[C@@H](N6C7=C(C(=NC=N7)N)N=C6)[C@@H]([C@@H]5OP([O-])([O-])=O)O)(=O)[O-])(=O)[O-])(C)C)O)=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H]([C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@@]3(O)CC[C@]12C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@@H]([C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C)C(=O)CC(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('C[C@H](CCCC(C)C(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12)[C@H]1CC[C@H]2[C@@H]3CCC4=CC(=O)CC[C@]4(C)[C@H]3CC[C@]12C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('[C@@H]1(N2C3=C(C(=NC=N3)N)N=C2)O[C@H](COP(OP(OCC(C)([C@H](C(NCCC(NCCSC(=O)CC(C(C)C)=O)=O)=O)O)C)(=O)[O-])(=O)[O-])[C@H]([C@H]1O)OP([O-])([O-])=O', "
               "'Molecule is a short chain fatty acyl-CoA(4-)'), "
               "('CC1=CC(=O)[C@H](CC(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)[C@@H]2OP([O-])([O-])=O)n2cnc3c(N)ncnc23)C1(C)C', "
               "'Molecule is a short chain fatty acyl-CoA(4-)')]\n"
               'False negatives: '
               "[('C[C@H](O)CC(=O)SCCNC(=O)CCNC(=O)[C@H](O)C(C)(C)COP([O-])(=O)OP([O-])(=O)OC[C@H]1O[C@H]([C@H](O)[C@@H]1OP([O-])([O-])=O)n1cnc2c(N)ncnc12', "
               "'Short chain fatty acyl group not found')]",
    'attempt': 2,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 1,
    'num_false_positives': 100,
    'num_true_negatives': 90882,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.009900990099009901,
    'recall': 1.0,
    'f1': 0.0196078431372549,
    'accuracy': 0.9989008935735247}


"""
Classifies: CHEBI:24838 inorganic phosphate
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_inorganic_phosphate(smiles: str):
    """
    Determines if a molecule is an inorganic phosphate (contains phosphate but no carbon).

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if molecule is an inorganic phosphate, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check if molecule contains a phosphate group
    has_phosphate = False
    for atom in mol.GetAtoms():
        if atom.GetSymbol() == 'P':
            neighbors = atom.GetNeighbors()
            oxygens = [n for n in neighbors if n.GetSymbol() == 'O']
            if len(oxygens) >= 3:
                has_phosphate = True
                break

    if not has_phosphate:
        return False, "No phosphate group found"

    # Check if molecule contains carbon atoms
    has_carbon = any(atom.GetSymbol() == 'C' for atom in mol.GetAtoms())
    if has_carbon:
        return False, "Molecule contains carbon atoms"

    return True, "Inorganic phosphate"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:24838',
                          'name': 'inorganic phosphate',
                          'definition': 'Any  phosphate that contains no '
                                        'carbon atom.',
                          'parents': ['CHEBI:24835', 'CHEBI:26020']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': '\n'
               'Attempt failed: F1 score of 0 is too low.\n'
               'True positives: []\n'
               "False positives: [('P(N)(=N)=N', 'Inorganic phosphate'), "
               "('NP(O)O', 'Inorganic phosphate'), ('NP(O)(O)=O', 'Inorganic "
               "phosphate'), ('[H]P([H])P([H])P([H])[H]', 'Inorganic "
               "phosphate')]\n"
               "False negatives: [('[Na+].OP(O)([O-])=O', 'No phosphate group "
               "found')]",
    'attempt': 1,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 1,
    'num_false_positives': 87,
    'num_true_negatives': 183832,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.011363636363636364,
    'recall': 1.0,
    'f1': 0.02247191011235955,
    'accuracy': 0.999526968247064}


"""
Classifies: CHEBI:26493 quinic acid
"""
from rdkit import Chem
from rdkit.Chem import AllChem

def is_quinic_acid(smiles: str):
    """
    Determines if a molecule is a quinic acid derivative.
    
    Args:
        smiles (str): SMILES string of the molecule
        
    Returns:
        bool: True if molecule is a quinic acid, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return None, "Invalid SMILES string"

    # Check for presence of core cyclohexane ring
    rings = mol.GetRingInfo()
    if not any(len(ring) == 6 for ring in rings.AtomRings()):
        return False, "No 6-membered ring found"
    
    # Check for carboxylic acid group
    carboxylic_acid_pattern = Chem.MolFromSmarts('C(=O)O')
    if not mol.HasSubstructMatch(carboxylic_acid_pattern):
        return False, "No carboxylic acid group found"

    # Pattern for quinic acid core structure
    # Cyclohexane ring with carboxylic acid and multiple OH groups
    quinic_core = Chem.MolFromSmarts('C1CC(C(=O)O)CC(O)C1')
    
    if mol.HasSubstructMatch(quinic_core):
        # Count hydroxyl groups
        oh_pattern = Chem.MolFromSmarts('[OH]')
        oh_matches = mol.GetSubstructMatches(oh_pattern)
        oh_count = len(oh_matches) if oh_matches else 0
        
        # Check for ester derivatives
        ester_pattern = Chem.MolFromSmarts('C(=O)OC')
        ester_matches = mol.GetSubstructMatches(ester_pattern)
        ester_count = len(ester_matches) if ester_matches else 0
        
        if ester_count > 0:
            return True, f"Quinic acid ester derivative with {oh_count} hydroxyl groups and {ester_count} ester groups"
        else:
            return True, f"Quinic acid with {oh_count} hydroxyl groups"
            
    return False, "Does not match quinic acid structure"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:26493',
                          'name': 'quinic acid',
                          'definition': 'A cyclitol carboxylic acid.',
                          'parents': ['CHEBI:36123']},
    'config': {   'llm_model_name': 'lbl/claude-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': 'Attempt failed: Python argument types in\n'
               '    Mol.GetSubstructMatches(Mol, NoneType)\n'
               'did not match C++ signature:\n'
               '    GetSubstructMatches(RDKit::ROMol self, RDKit::MolBundle '
               'query, RDKit::SubstructMatchParameters params)\n'
               '    GetSubstructMatches(RDKit::ROMol self, RDKit::ROMol query, '
               'RDKit::SubstructMatchParameters params)\n'
               '    GetSubstructMatches(RDKit::ROMol self, RDKit::MolBundle '
               'query, bool uniquify=True, bool useChirality=False, bool '
               'useQueryQueryMatches=False, unsigned int maxMatches=1000)\n'
               '    GetSubstructMatches(RDKit::ROMol self, RDKit::ROMol query, '
               'bool uniquify=True, bool useChirality=False, bool '
               'useQueryQueryMatches=False, unsigned int maxMatches=1000)',
    'attempt': 1,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 3,
    'num_false_positives': 100,
    'num_true_negatives': 37130,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.02912621359223301,
    'recall': 1.0,
    'f1': 0.05660377358490566,
    'accuracy': 0.99731421051218}


"""
Classifies: CHEBI:33249 organyl group
"""
from rdkit import Chem
from rdkit.Chem import Descriptors

def is_organyl_group(smiles: str):
    """
    Determines if a given SMILES string represents an organyl group.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if the molecule is an organyl group, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check for presence of a free valence at a carbon atom
    for atom in mol.GetAtoms():
        if atom.GetSymbol() == 'C' and atom.GetTotalNumHs() < atom.GetTotalValence():
            # Found a carbon atom with a free valence
            break
    else:
        return False, "No free valence found at a carbon atom"

    # Check if the molecule is organic
    if not all(atom.GetSymbol() in ['C', 'H', 'O', 'N', 'S', 'P', 'F', 'Cl', 'Br', 'I'] for atom in mol.GetAtoms()):
        return False, "Molecule contains inorganic atoms"

    return True, "Organyl group"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:33249',
                          'name': 'organyl group',
                          'definition': 'Any organic substituent group, '
                                        'regardless of functional type, having '
                                        'one free valence at a carbon atom.',
                          'parents': ['CHEBI:51447']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': "Attempt failed: 'FragmentMatcher' object has no attribute "
               "'AddFragmentSmarts'",
    'attempt': 1,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 1,
    'num_false_positives': 100,
    'num_true_negatives': 5,
    'num_false_negatives': 20,
    'num_negatives': None,
    'precision': 0.009900990099009901,
    'recall': 0.047619047619047616,
    'f1': 0.01639344262295082,
    'accuracy': 0.047619047619047616}


"""
Classifies: CHEBI:3098 bile acid
"""
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdMolDescriptors

def is_bile_acid(smiles: str):
    """
    Determines if a molecule is a bile acid.

    Bile acids are defined as hydroxy-5beta-cholanic acids occurring in bile, where they are present as the
    sodium salts of their amides with glycine or taurine. In mammals, bile acids almost invariably have
    a 5beta-configuration.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if the molecule is a bile acid, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check for the 5beta-cholanic acid scaffold
    ring_info = mol.GetRingInfo()
    cholanic_rings = []
    for ring in ring_info.AtomRings():
        if len(ring) == 17:
            atoms = [mol.GetAtomWithIdx(i) for i in ring]
            is_cholanic = all(atom.GetSymbol() == 'C' for atom in atoms)
            if is_cholanic:
                cholanic_rings.append(ring)

    if not cholanic_rings:
        return False, "No cholanic acid scaffold found"

    # Check for hydroxy groups
    hydroxy_count = Chem.Crippen.MolLogP(mol) < 4.0  # Empirical threshold based on logP
    if not hydroxy_count:
        return False, "No hydroxy groups found"

    # Check for 5beta-configuration
    ring_atoms = set(cholanic_rings[0])
    ring_atoms_sorted = sorted(ring_atoms, key=lambda x: mol.GetAtomWithIdx(x).GetDegree(), reverse=True)
    ring_atom_idx = ring_atoms_sorted[4]  # 5th highest degree atom is the ring junction
    ring_atom = mol.GetAtomWithIdx(ring_atom_idx)

    if ring_atom.GetHybridization() != Chem.HybridizationType.SP3:
        return False, "Ring junction atom is not sp3 hybridized"

    ring_atom_neighbors = [mol.GetAtomWithIdx(neighbor_idx) for neighbor_idx in ring_atom.GetNeighbors()]
    hydrogens = [neighbor for neighbor in ring_atom_neighbors if neighbor.GetSymbol() == 'H']

    if len(hydrogens) != 1:
        return False, "Ring junction atom does not have a single hydrogen"

    if hydrogens[0].GetIsotope():  # Check for deuterium
        return False, "Ring junction hydrogen is deuterated"

    return True, "Molecule is a bile acid"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:3098',
                          'name': 'bile acid',
                          'definition': 'Any member of a group of '
                                        'hydroxy-5beta-cholanic acids occuring '
                                        'in bile, where they are present as '
                                        'the sodium salts of their amides with '
                                        'glycine or taurine. In mammals bile '
                                        'acids almost invariably have '
                                        '5beta-configuration.',
                          'parents': ['CHEBI:138366', 'CHEBI:24663']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': False,
    'best': True,
    'error': 'Python argument types in\n'
             '    Mol.GetAtomWithIdx(Mol, Atom)\n'
             'did not match C++ signature:\n'
             '    GetAtomWithIdx(RDKit::ROMol {lvalue} self, unsigned int idx)',
    'stdout': '',
    'num_true_positives': 0,
    'num_false_positives': 0,
    'num_true_negatives': 0,
    'num_false_negatives': 0,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0.0,
    'accuracy': None}


"""
Classifies: CHEBI:21368 L-ornithine derivative
"""
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import Descriptors
from rdkit.Chem import rdMolDescriptors

def is_L_ornithine_derivative(smiles: str):
    """
    Determines if a molecule is an L-ornithine derivative.

    An L-ornithine derivative is defined as resulting from a reaction of L-ornithine at the
    amino group, the carboxy group or the side-chain amino group, or from the replacement of
    any hydrogen of L-ornithine by a heteroatom.

    Args:
        smiles (str): SMILES string of the molecule

    Returns:
        bool: True if the molecule is an L-ornithine derivative, False otherwise
        str: Reason for classification
    """
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return False, "Invalid SMILES string"

    # Check if the molecule contains the L-ornithine backbone
    ornithine_smarts = '[NH2][CH]([CH2][CH2][CH2][CH](N)[CH](O)=O)O'
    ornithine_match = mol.GetSubstructMatches(Chem.MolFromSmarts(ornithine_smarts))

    if not ornithine_match:
        return False, "The molecule does not contain the L-ornithine backbone"

    # Check for modifications at the amino, carboxy, or side-chain amino groups
    modified_groups = []
    for atom_idx in ornithine_match[0]:
        atom = mol.GetAtomWithIdx(atom_idx)
        if atom.GetSymbol() == 'N':
            if atom.GetTotalNumHs() < 2:
                modified_groups.append('amino')
        elif atom.GetSymbol() == 'O':
            if atom.GetTotalNumHs() < 1:
                modified_groups.append('carboxy')
        elif atom.GetSymbol() == 'C':
            neighbors = [mol.GetAtomWithIdx(neighbor_idx) for neighbor_idx in atom.GetNeighbors()]
            if any(neighbor.GetSymbol() == 'N' and neighbor.GetTotalNumHs() < 2 for neighbor in neighbors):
                modified_groups.append('side-chain amino')

    # Check for replacement of hydrogens with heteroatoms
    heteroatom_replacement = False
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() != 1 and atom.GetAtomicNum() != 6 and atom.GetAtomicNum() != 7 and atom.GetAtomicNum() != 8:
            heteroatom_replacement = True
            break

    if modified_groups or heteroatom_replacement:
        reason = "The molecule is an L-ornithine derivative with modifications: "
        if modified_groups:
            reason += ', '.join(modified_groups)
        if heteroatom_replacement:
            if modified_groups:
                reason += ', '
            reason += "heteroatom replacement"
        return True, reason
    else:
        return False, "The molecule is L-ornithine itself, not a derivative"


__metadata__ = {   'chemical_class': {   'id': 'CHEBI:21368',
                          'name': 'L-ornithine derivative',
                          'definition': 'An ornithine derivative resulting '
                                        'from reaction of L-ornithine at the '
                                        'amino group, the carboxy group or the '
                                        'side-chain amino group, or from the '
                                        'replacement of any hydrogen of '
                                        'L-ornithine by a heteroatom.',
                          'parents': ['CHEBI:25718']},
    'config': {   'llm_model_name': 'claude-3-sonnet',
                  'f1_threshold': 0.0,
                  'max_attempts': 5,
                  'max_negative_to_test': None,
                  'max_positive_in_prompt': 50,
                  'max_negative_in_prompt': 20,
                  'max_instances_in_prompt': 100,
                  'test_proportion': 0.1},
    'message': None,
    'attempt': 0,
    'success': True,
    'best': True,
    'error': '',
    'stdout': None,
    'num_true_positives': 0,
    'num_false_positives': 0,
    'num_true_negatives': 183918,
    'num_false_negatives': 1,
    'num_negatives': None,
    'precision': 0.0,
    'recall': 0.0,
    'f1': 0,
    'accuracy': 0.9999945628238518}